import uuid

from django.db import models
from django.conf import settings
from django.urls import reverse
from django.utils import timezone

CONTRACTOR_DIR = getattr(settings, 'CONTRACTOR_DIR', 'contractor')
CONTRACTOR_URL = getattr(settings, 'CONTRACTOR_URL', settings.MEDIA_URL)


def get_url_path(slug, version, filename):
    return '/'.join([CONTRACTOR_DIR, slug, str(version), filename])


def is_full_url(line):
    return line.startswith(('https://', 'http://'))


class Contract(models.Model):
    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)

    created = models.DateTimeField(default=timezone.now)
    updated = models.DateTimeField(default=timezone.now)

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        related_name='+',
        on_delete=models.SET_NULL
    )
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        related_name='+',
        on_delete=models.SET_NULL
    )
    token = models.UUIDField(default=uuid.uuid4)

    version = models.IntegerField(default=0)
    source_url = models.URLField()

    files = models.TextField(blank=True)

    webhook_log = models.TextField(blank=True)

    class Meta:
        verbose_name = 'contract'
        verbose_name_plural = 'contracts'

    def __str__(self):
        return self.name

    def get_absolute_url(self, filename=''):
        return CONTRACTOR_URL + get_url_path(
            self.slug,
            self.version,
            filename
        )

    def get_webhook_url(self):
        return reverse('contractor:webhook', kwargs={'token': self.token})

    def get_source_files(self):
        for line in self.files.splitlines():
            line = line.strip()
            if not line:
                continue
            if is_full_url(line):
                continue
            else:
                yield (line, self.source_url + line)

    def get_file_path(self, filename=''):
        return get_url_path(self.slug, self.version, filename)

    def get_file_url(self, filename=''):
        if is_full_url(filename):
            return filename

        path = get_url_path(self.slug, self.version, filename)
        return CONTRACTOR_URL + path

    def get_file_urls(self, files=None):
        if files is None:
            files = self.files
        for line in files.splitlines():
            line = line.strip()
            if not line:
                continue
            yield self.get_file_url(line)


try:
    from .cms_models import ContractWorkPlugin  # noqa
except ImportError:
    # Ignore errors if Django CMS is not installed
    pass



from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.db.models import F
from django.utils import timezone

from celery import shared_task
import requests

from .models import Contract


@shared_task(name='fetch_contract_result')
def fetch_contract_result(contract_id):
    try:
        contract = Contract.objects.get(id=contract_id)
    except Contract.DoesNotExist:
        return

    Contract.objects.filter(id=contract_id).update(
        version=F('version') + 1,
        updated=timezone.now()
    )
    contract.version += 1

    for filename, url in contract.get_source_files():
        download_file(contract, filename, url)


def download_file(contract, filename, url):
    path = contract.get_file_path(filename)
    response = requests.get(url)
    # FIXME: returned path is not stored, assumed to be the same
    default_storage.save(path, ContentFile(response.content))



# Generated by Django 2.2.4 on 2020-04-15 15:06

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0022_auto_20180620_1551'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Contract',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(unique=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('updated', models.DateTimeField(default=django.utils.timezone.now)),
                ('token', models.UUIDField(default=uuid.uuid4)),
                ('version', models.IntegerField(default=0)),
                ('source_url', models.URLField()),
                ('files', models.TextField(blank=True)),
                ('webhook_log', models.TextField(blank=True)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'contract',
                'verbose_name_plural': 'contracts',
            },
        ),
        migrations.CreateModel(
            name='ContractWorkPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='contractor_contractworkplugin', serialize=False, to='cms.CMSPlugin')),
                ('javascript', models.TextField(blank=True)),
                ('styles', models.TextField(blank=True)),
                ('version', models.IntegerField(default=0)),
                ('xpath', models.CharField(blank=True, max_length=255)),
                ('html', models.TextField(blank=True)),
                ('contract', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contractor.Contract')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import ContractWorkPlugin


@plugin_pool.register_plugin
class ContractWorkCMSPlugin(CMSPluginBase):
    """
    Plugin for including a selection of entries
    """
    module = 'Contract'
    name = 'Render contract work'
    model = ContractWorkPlugin
    render_template = 'contractor/render.html'
    text_enabled = True

    def render(self, context, instance, placeholder):
        """
        Update the context with plugin's data
        """
        context = super().render(context, instance, placeholder)
        context['instance'] = instance
        context['html'] = instance.get_html()
        context['path'] = instance.contract.get_file_url()
        return context



import logging

from django.core.files.storage import default_storage

import lxml.html

logger = logging.getLogger(__name__)


def extract_html(contract, xpath):
    filename, xpath = xpath.split('#', 1)
    path = contract.get_file_path(filename)
    try:
        content = default_storage.open(path).read()
        root = lxml.html.fromstring(content.decode('utf-8'))
    except Exception as e:
        logger.exception(e)
        return ''

    # Convert relative img paths to full media URLs
    selections = root.xpath(xpath)
    for selection in selections:
        for img in selection.xpath('//img'):
            src = img.attrib['src']
            if not src.startswith(('http://', 'https://', '/')):
                img.attrib['src'] = contract.get_file_url(src)

    try:
        return '\n'.join(
            lxml.html.tostring(s, pretty_print=True).decode('utf-8')
            for s in selections
        )
    except Exception as e:
        logger.exception(e)
    return ''



from django.contrib import admin
from django.utils import timezone

from .models import Contract


class ContractAdmin(admin.ModelAdmin):
    list_display = (
        'name', 'created', 'updated',
        'created_by', 'updated_by', 'version'
    )
    readonly_fields = (
        'created', 'updated', 'created_by', 'updated_by',
        'version', 'webhook_log'
    )
    prepopulated_fields = {'slug': ('name',)}
    save_on_top = True
    ordering = ('-updated',)
    date_hierarchy = 'updated'
    actions = ['manual_update']

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created = timezone.now()
            obj.created_by = request.user
        obj.updated = timezone.now()
        obj.updated_by = request.user
        super().save_model(request, obj, form, change)

    def manual_update(self, request, queryset):
        from .tasks import fetch_contract_result

        for contract in queryset:
            fetch_contract_result.delay(contract.pk)

        self.message_user(request, "Contract update triggered.")


admin.site.register(Contract, ContractAdmin)



from django.db import models

from cms.models.pluginmodel import CMSPlugin


class ContractWorkPlugin(CMSPlugin):
    """
    CMS Plugin for displaying latest investigations
    """

    contract = models.ForeignKey(
        'Contract', on_delete=models.CASCADE
    )

    javascript = models.TextField(blank=True)
    styles = models.TextField(blank=True)

    version = models.IntegerField(default=0)
    xpath = models.CharField(max_length=255, blank=True)
    html = models.TextField(blank=True)

    def __str__(self):
        return self.contract.name

    def get_html(self):
        if self.contract.version == self.version:
            return self.html

        if not self.xpath:
            return ''

        from .utils import extract_html

        self.html = extract_html(self.contract, self.xpath)
        self.version = self.contract.version
        self.save()

        return self.html

    def get_js(self):
        return list(self.contract.get_file_urls(self.javascript))

    def get_css(self):
        return list(self.contract.get_file_urls(self.styles))



from django.http import JsonResponse
from django.views.decorators.http import require_POST
from django.views.decorators.csrf import csrf_exempt

from .models import Contract
from .tasks import fetch_contract_result


@csrf_exempt
@require_POST
def webhook(request, token):
    try:
        contract = Contract.objects.get(token=token)
    except Contract.DoesNotExist:
        return JsonResponse({})

    contract.webhook_log = request.body.decode('utf-8')
    contract.save()

    fetch_contract_result.delay(contract.id)

    return JsonResponse({})



#!/usr/bin/env python

import codecs
import re
import os

from setuptools import setup, find_packages


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="django-contractor",
    version=find_version("contractor", "__init__.py"),
    url='https://github.com/stefanw/django-contractor',
    license='MIT',
    description="Fetch media files that were build elsewhere",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Stefan Wehrmeyer',
    author_email='mail@stefanwehrmeyer.com',
    packages=find_packages(exclude=['tests*']),
    install_requires=[
        'Django', 'requests', 'celery', 'lxml'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP'
    ]
)
